/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"strings"
	"sync"

	"github.com/google/blueprint"

	"bob/utils"
)

const (
	STATIC_LIBRARY = iota
	SHARED_LIBRARY = iota
	EXECUTABLE     = iota
)

type AndroidMkGenerator struct{}

var androidLock sync.Mutex

func writeIfChanged(filename string, text string) {
	mustWrite := true

	// If any errors occur trying to determine the state of the existing file,
	// just write the new file
	fileinfo, err := os.Stat(filename)
	if err == nil {
		if fileinfo.Size() == int64(len(text)) {
			current, err := ioutil.ReadFile(filename)
			if err == nil {
				if string(current) == text {
					// No need to write
					mustWrite = false
				}
			}
		}
	}

	if mustWrite {
		file, err := os.OpenFile(filename, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0600)
		if err != nil {
			panic(err)
		}

		file.WriteString(text)
		file.Close()
	}
}

func androidMkWriteString(ctx blueprint.ModuleContext, name string, text string) {
	filename := filepath.Join(builddir, name+".inc")
	writeIfChanged(filename, text)
}

func CopyGeneratedSources(ctx blueprint.ModuleContext) {

}

func newlineSeparatedList(list []string) string {
	return " \\\n    " + strings.Join(list, " \\\n    ") + "\n"
}

// This flag is a machine specific option
func machineSpecificFlag(s string) bool {
	return strings.HasPrefix(s, "-m")
}

// This flag selects the compiler standard
func compilerStandard(s string) bool {
	return strings.HasPrefix(s, "-std=")
}

// Identify whether a compilation flag should be used on android
//
// The Android build system should set machine specific flags (so it
// can do multi-arch builds) and compiler standard, so filter these
// out from module properties.
func moduleCompileFlags(s string) bool {
	return !(machineSpecificFlag(s) || compilerStandard(s))
}

// Identify whether a link flag should be used on android
//
// The Android build system should set machine specific flags (so it
// can do multi-arch builds), so filter these out from module
// properties.
func moduleLinkFlags(s string) bool {
	return !machineSpecificFlag(s)
}

// If we see any of there libraries in LDLIBS as -lxxx, we'll replace them with the android library object
var androidSharedLibs = [...]string{
	"libbinder",
	"libc++",
	"libcutils",
	"libgui",
	"libhardware",
	"libion",
	"liblog",
	"libnativewindow",
	"libsync",
	"libui",
	"libutils"}

var androidStaticLibs = [...]string{
	"libarect"}

var androidHeaderLibs = [...]string{
	"libcutils_headers",
	"libgui_headers",
	"libhardware_headers",
	"liblog_headers",
	"libnativebase_headers",
	"libsystem_headers",
	"libui_headers"}

func noAndroidLdlibs(s string) bool {
	// An android module should have a way to add android libraries to its LOCAL_SHARED_LIBRARIES,
	// LOCAL_HEADER_LIBRARIES and LOCAL_STATIC_LIBRARIES. The way we do this is via ldlibs, if a
	// module has a definition in ldlibs which doesn't pass this predicate, it is this kind of
	// library. In this case, module is removed from LOCAL_LDLIBS and instead added to either,
	// LOCAL_SHARED_LIBS, LOCAL_HEADER_LIBS or LOCAL_STATIC_LIBS.
	if strings.HasPrefix(s, "android.") {
		return false
	}
	for _, lib := range androidSharedLibs {
		if s[2:] == lib[3:] {
			// Don't include android shared lib
			return false
		}
	}
	for _, lib := range androidHeaderLibs {
		if s[2:] == lib[3:] {
			// Don't include android header lib
			return false
		}
	}
	for _, lib := range androidStaticLibs {
		if s[2:] == lib[3:] {
			// Don't include android static lib
			return false
		}
	}
	return true
}

var (
	classes = []string{
		"STATIC_LIBRARIES",
		"SHARED_LIBRARIES",
		"EXECUTABLES",
	}

	rulePrefix = map[string]string{
		TARGET_TAG: "BUILD_",
		HOST_TAG:   "BUILD_HOST_",
	}

	ruleSuffix = []string{
		"STATIC_LIBRARY",
		"SHARED_LIBRARY",
		"EXECUTABLE",
	}
)

func specifyCompilerStandard(varname string, flags []string) string {
	// Look for the flag setting compiler standard
	line := ""
	stdList := utils.Filter(flags, compilerStandard)
	if len(stdList) > 0 {
		// Use last definition only
		std := strings.TrimPrefix(stdList[len(stdList)-1], "-std=")
		line += varname + ":=" + std + "\n"
	}
	return line
}

func (m *Library) GenerateBuildAction(subtype int, ctx blueprint.ModuleContext) {
	if m.Properties.Build_wrapper != nil {
		panic(errors.New("build_wrapper not supported on Android"))
	}

	text := "##########################\ninclude $(CLEAR_VARS)\n\n"
	text += "LOCAL_MODULE:=" + m.AltName() + "\n"
	text += "LOCAL_MODULE_CLASS:=" + classes[subtype] + "\n\n"

	// The order we want is  local_include_dirs, export_local_include_dirs,
	//                       include_dirs, export_include_dirs
	// This is because include and export include should be system headers
	includes := utils.PrefixDirs(m.Properties.Local_include_dirs, "$(LOCAL_PATH)")
	includes = append(includes, utils.PrefixDirs(m.Properties.Export_local_include_dirs, "$(LOCAL_PATH)")...)
	includes = append(includes, m.Properties.Include_dirs...)
	includes = append(includes, m.Properties.Export_include_dirs...)

	export_include_dirs := m.Properties.Export_include_dirs
	export_include_dirs = append(export_include_dirs, utils.PrefixDirs(m.Properties.Export_local_include_dirs, "$(LOCAL_PATH)")...)

	// Handle generated headers
	if len(m.Properties.Generated_headers) > 0 {
		header_dirs, header_outputs := m.GetGeneratedHeaders(ctx)
		includes = append(includes, header_dirs...)
		headers := strings.Join(header_outputs, " ")

		text += "LOCAL_ADDITIONAL_DEPENDENCIES := " + headers + "\n\n"
	}

	// Handle generated sources
	for _, module := range m.Properties.Generated_sources {
		// LOCAL_GENERATED_SOURCES is used to name target generated as
		// part of this module which we also link into a library. The
		// generated sources are automatically added to the
		// library. Unfortunately, we've generated the sources in a
		// separate module...
		//
		// To compile a file generated in another module we could try
		// and explicitly list the file in that modules directory in
		// LOCAL_SRCS. However Android make won't work with LOCAL_SRCS
		// outside the source tree, so we can't do that.
		//
		// Therefore we use LOCAL_GENERATED_SOURCES and copy the files
		// generated in the other module into this module.
		//
		// An alternative would be to avoid the separate source
		// generation module and do it as part of this module.

		sources := "$(" + module + "_OUTPUTS)"
		sources_dir := "$(" + module + "_GEN_DIR)"

		local_source_expr := "$(subst " + sources_dir + ", $(local-generated-sources-dir), " + sources + ")"
		local_sources := "$(" + m.AltName() + "_" + module + "_SRCS)"

		text += m.AltName() + "_" + module + "_SRCS:=" + local_source_expr + "\n"
		text += "LOCAL_GENERATED_SOURCES+=" + local_sources + "\n"

		// Copy rule. Use a static pattern to avoid running the command for each file
		text += local_sources + ": $(local-generated-sources-dir)" + "/%: " + sources_dir + "/%\n"
		text += "\tcp $< $@\n\n"
	}

	if getConfig(ctx).Properties.GetBool("toolchain_clang") {
		text += "LOCAL_CLANG := true\n"
	} else {
		text += "LOCAL_CLANG := false\n"
	}
	srcs := append(m.Properties.GetSrcs(ctx), m.Properties.Build.SourceProps.Specials...)
	text += "LOCAL_SRC_FILES := " + strings.Join(srcs, " ") + "\n"

	text += "LOCAL_C_INCLUDES := " + strings.Join(includes, " ") + "\n"
	cflags_list := append(m.Properties.Cflags, m.Properties.Export_cflags...)
	_, exported_cflags := m.GetExportedVariables(ctx)
	cflags_list = append(cflags_list, exported_cflags...)
	text += "LOCAL_CFLAGS := " + strings.Join(utils.Filter(cflags_list, moduleCompileFlags), " ") + "\n"
	text += "LOCAL_CPPFLAGS := " + strings.Join(utils.Filter(m.Properties.Cxxflags, moduleCompileFlags), " ") + "\n"
	text += "LOCAL_CONLYFLAGS := " + strings.Join(utils.Filter(m.Properties.Conlyflags, moduleCompileFlags), " ") + "\n"

	// Setup module C/C++ standard if requested. Note that this only affects Android O and later.
	text += specifyCompilerStandard("LOCAL_C_STD", append(cflags_list, m.Properties.Conlyflags...))
	text += specifyCompilerStandard("LOCAL_CPP_STD", append(cflags_list, m.Properties.Cxxflags...))

	// Check for android libraries in ldlibs, and add to
	// shared, header or static libs instead of ldlibs.
	// This means android will add the appropriate
	// includes and build the right things.
	android_shared_libs := []string{}
	android_header_libs := []string{}
	android_static_libs := []string{}
	if len(m.Properties.Ldlibs) > 0 {
		// The following code is similar to filter, but we are
		// transforming the entries at the same time.
		for _, lib := range m.Properties.Ldlibs {
			if strings.HasPrefix(lib, "android.") {
				android_shared_libs = append(android_shared_libs, lib)
				continue
			}
			for _, lib2 := range androidSharedLibs {
				if lib[2:] == lib2[3:] {
					android_shared_libs = append(android_shared_libs, lib2)
					break
				}
			}
			for _, lib2 := range androidHeaderLibs {
				if lib[2:] == lib2[3:] {
					android_header_libs = append(android_header_libs, lib2)
					break
				}
			}
			for _, lib2 := range androidStaticLibs {
				if lib[2:] == lib2[3:] {
					android_static_libs = append(android_static_libs, lib2)
					break
				}
			}
		}
		// Filter out the android libs from ldlibs
		m.Properties.Ldlibs = utils.Filter(m.Properties.Ldlibs, noAndroidLdlibs)
	}
	// Similar processing of export ldlibs to ldlibs.
	if len(m.Properties.Export_ldlibs) > 0 {
		// The following code is similar to filter, but we are
		// transforming the entries at the same time.
		for _, lib := range m.Properties.Export_ldlibs {
			for _, lib2 := range androidSharedLibs {
				if lib[2:] == lib2[3:] {
					android_shared_libs = append(android_shared_libs, lib2)
					break
				}
			}
			for _, lib2 := range androidHeaderLibs {
				if lib[2:] == lib2[3:] {
					android_header_libs = append(android_header_libs, lib2)
					break
				}
			}
			for _, lib2 := range androidStaticLibs {
				if lib[2:] == lib2[3:] {
					android_static_libs = append(android_static_libs, lib2)
					break
				}
			}
		}
	}

	// convert Shared_libs, Resolved_static_libs, Whole_static_libs to
	// Android module names rather than Bob module names
	shared_libs := []string{}
	static_libs := []string{}
	whole_static_libs := []string{}
	for _, mod := range m.Properties.Shared_libs {
		shared_libs = append(shared_libs, androidModuleName(mod))
	}
	for _, mod := range m.Properties.Resolved_static_libs {
		static_libs = append(static_libs, androidModuleName(mod))
	}
	for _, mod := range m.Properties.Whole_static_libs {
		whole_static_libs = append(whole_static_libs, androidModuleName(mod))
	}

	text += "LOCAL_SHARED_LIBRARIES := " + strings.Join(android_shared_libs, " ") + " " +
		strings.Join(shared_libs, " ") +
		" liblog libc++\n"
	text += "LOCAL_STATIC_LIBRARIES := " + strings.Join(static_libs, " ") + "\n"
	text += "LOCAL_WHOLE_STATIC_LIBRARIES := " + strings.Join(whole_static_libs, " ") + "\n"

	text += "LOCAL_MODULE_TAGS := " + strings.Join(m.Properties.Tags, " ") + "\n"
	text += "LOCAL_EXPORT_C_INCLUDE_DIRS := " + strings.Join(export_include_dirs, " ") + "\n"
	if m.Properties.Owner != "" {
		text += "LOCAL_MODULE_OWNER := " + m.Properties.Owner + "\n"
		text += "LOCAL_PROPRIETARY_MODULE := true\n"
	}

	if len(android_header_libs) > 0 {
		text += "LOCAL_HEADER_LIBRARIES := " + strings.Join(android_header_libs, " ") + "\n"
	}

	if len(android_static_libs) > 0 {
		text += "LOCAL_STATIC_LIBRARIES += " + strings.Join(android_static_libs, " ") + "\n"
	}

	// Can't see a way to wrap a particular library in -Wl in link flags on android, so specify
	// -Wl,--copy-dt-needed-entries across the lot
	hasForwardingLib := false
	copydtneeded := ""
	ctx.VisitDirectDepsIf(
		func(p blueprint.Module) bool { return ctx.OtherModuleDependencyTag(p) == SharedDepTag },
		func(p blueprint.Module) {
			if sl, ok := p.(*SharedLibrary); ok {
				b := sl.Build()
				if b.IsForwardingSharedLibrary() {
					hasForwardingLib = true
				}
			} else if _, ok := p.(*GenerateSharedLibrary); ok {
				// Generated forwarding lib not supported
			} else {
				panic(errors.New(ctx.OtherModuleName(p) + " is not a shared library"))
			}
		})
	if hasForwardingLib {
		copydtneeded = "-Wl,--copy-dt-needed-entries"
	}

	// Handle installation
	installGroupPath, ok := getInstallGroupPath(ctx)

	// Only setup multilib for target modules.
	// Normally this should only apply to target libraries, but we
	// also do multilib target binaries to allow creation of test
	// binaries in both modes.
	// All test binaries will be installable.
	isMultiLib := (m.Properties.Target_type == TARGET_TAG) &&
		((subtype == SHARED_LIBRARY) || (subtype == STATIC_LIBRARY) || ok)

	if ok {
		text += "LOCAL_MODULE_RELATIVE_PATH:=" + m.Properties.Relative_install_path + "\n"
		if m.Properties.Post_install_cmd != "" {
			// Setup args like we do for bob_generated_*
			args := map[string]string{}
			args["bob_config"] = "$(BOB_ANDROIDMK_DIR)/" + configName
			args["tool"] = filepath.Join("$(LOCAL_PATH)", ctx.ModuleDir(), m.Properties.Post_install_tool)
			args["out"] = "$(LOCAL_INSTALLED_MODULE)"

			// We can't use target specific variables in make due to
			// the way LOCAL_POST_INSTALL_CMD is
			// implemented. Therefore expand all variable use here.
			cmd := m.Properties.Post_install_cmd
			for key, value := range args {
				cmd = strings.Replace(cmd, "${"+key+"}", value, -1)
			}

			// Intentionally using a recursively expanded variable. We
			// don't want LOCAL_INSTALLED_MODULE expanded now, but
			// when it is used in base_rules.mk
			text += "LOCAL_POST_INSTALL_CMD=" + cmd + "\n"
		}

		if subtype == EXECUTABLE {
			if isMultiLib {
				// For executables we need to be clear about where to
				// install both 32 and 64 bit versions of the
				// binaries.
				// LOCAL_UNSTRIPPED_PATH does not need to be set
				text += "LOCAL_MODULE_PATH_32:=" + installGroupPath + "\n"
				text += "LOCAL_MODULE_PATH_64:=" + installGroupPath + "64\n"
			} else {
				// When LOCAL_MODULE_PATH is specified, you need to
				// specify LOCAL_UNSTRIPPED_PATH too
				text += "LOCAL_MODULE_PATH:=" + installGroupPath + "\n"

				if m.Properties.Target_type == TARGET_TAG {
					// Unstripped executables only generated for target
					text += "LOCAL_UNSTRIPPED_PATH:=$(TARGET_OUT_EXECUTABLES_UNSTRIPPED)\n"
				}
			}
		} else {
			// You can't specify an explicit install dir for
			// libraries, you have to use
			// LOCAL_MODULE_RELATIVE_PATH
		}

		requiredModuleNames := m.GetInstallDepPhonyNames(ctx)
		if len(requiredModuleNames) > 0 {
			text += "LOCAL_REQUIRED_MODULES:=" + newlineSeparatedList(requiredModuleNames)
		}
	} else {
		if m.Properties.Target_type == TARGET_TAG {
			// Only disable installation of target libraries.
			// Host libraries need to be installed to be used by the build.
			text += "LOCAL_UNINSTALLABLE_MODULE:=true\n"
		}
	}

	if isMultiLib {
		text += "LOCAL_MULTILIB:=both\n"
		text += "LOCAL_LDFLAGS_32:=" + strings.Join(utils.Filter(m.Properties.Ldflags, moduleLinkFlags), " ") + copydtneeded + "\n"
	}
	text += "LOCAL_LDFLAGS:=" + strings.Join(utils.Filter(m.Properties.Ldflags, moduleLinkFlags), " ") + copydtneeded + "\n"

	if m.Properties.Target_type == TARGET_TAG {
		text += "LOCAL_LDLIBS := " + strings.Join(m.Properties.Ldlibs, " ") + "\n"
	} else {
		text += "LOCAL_LDLIBS_$(HOST_OS) := " + strings.Join(m.Properties.Ldlibs, " ") + "\n"
	}
	text += "\ninclude $(" + rulePrefix[m.Properties.Target_type] + ruleSuffix[subtype] + ")\n"

	androidMkWriteString(ctx, m.AltShortName(), text)
}

func (g *AndroidMkGenerator) StaticActions(m *StaticLibrary, ctx blueprint.ModuleContext) {
	if enabledAndRequired(m) {
		m.GenerateBuildAction(STATIC_LIBRARY, ctx)
	}
}

func (g *AndroidMkGenerator) SharedActions(m *SharedLibrary, ctx blueprint.ModuleContext) {
	if enabledAndRequired(m) {
		m.GenerateBuildAction(SHARED_LIBRARY, ctx)
	}
}

func (g *AndroidMkGenerator) BinaryActions(m *Binary, ctx blueprint.ModuleContext) {
	if enabledAndRequired(m) {
		m.GenerateBuildAction(EXECUTABLE, ctx)
	}
}

func (*AndroidMkGenerator) declareAlias(name string, srcs []string) string {
	text := "\ninclude $(CLEAR_VARS)\n\n"
	text += "LOCAL_MODULE := " + name + "\n"

	text += "LOCAL_REQUIRED_MODULES :=" + newlineSeparatedList(srcs)

	text += "\n.PHONY: " + name + "\n"
	text += name + ": $(LOCAL_REQUIRED_MODULES)\n\n"

	text += "include $(base_rules.mk)\n"

	return text
}

func (g *AndroidMkGenerator) AliasActions(m *Alias, ctx blueprint.ModuleContext) {
	text := g.declareAlias(m.Name(), m.Properties.Srcs)
	androidMkWriteString(ctx, m.Name(), text)
}

func pathToModuleName(path string) string {
	path = strings.Replace(path, "/", "__", -1)
	path = strings.Replace(path, ".", "_", -1)
	path = strings.Replace(path, "(", "_", -1)
	path = strings.Replace(path, "$", "_", -1)
	path = strings.Replace(path, ")", "_", -1)
	return path
}

func (g *AndroidMkGenerator) ResourceActions(m *Resource, ctx blueprint.ModuleContext) {
	if !enabledAndRequired(m) {
		return
	}

	installGroupPath, ok := getInstallGroupPath(ctx)
	if !ok {
		androidMkWriteString(ctx, m.AltShortName(), "")
		return
	}

	filesToInstall := m.FilesToInstall(ctx)
	requiredModuleNames := m.GetInstallDepPhonyNames(ctx)

	text := ""

	for _, file := range filesToInstall {
		moduleName := pathToModuleName(file)
		requiredModuleNames = append(requiredModuleNames, moduleName)

		text += "\ninclude $(CLEAR_VARS)\n\n"
		text += "LOCAL_MODULE := " + moduleName + "\n"
		text += "LOCAL_INSTALLED_MODULE_STEM := " + filepath.Base(file) + "\n"
		text += "LOCAL_MODULE_CLASS := ETC\n"
		text += "LOCAL_MODULE_PATH := " + installGroupPath + "\n"
		text += "LOCAL_MODULE_RELATIVE_PATH := " + m.Properties.Relative_install_path + "\n"
		text += "LOCAL_MODULE_TAGS := " + strings.Join(m.Properties.Tags, " ") + "\n"
		text += "LOCAL_SRC_FILES := " + file + "\n"
		text += "\ninclude $(BUILD_PREBUILT)\n"
	}

	text += g.declareAlias(m.Name(), requiredModuleNames)

	androidMkWriteString(ctx, m.AltShortName(), text)
}

func (g *AndroidMkGenerator) sourcePrefix() string {
	return "$(LOCAL_PATH)"
}

func (g *AndroidMkGenerator) buildDir() string {
	return "$(BOB_ANDROIDMK_DIR)"
}

func outputDirVarName(m *GenerateCommon) string {
	return m.Name() + "_GEN_DIR"
}

func (g *AndroidMkGenerator) sourceOutputDir(m *GenerateCommon) string {
	if m.Properties.Target != HOST_TAG {
		return filepath.Join("$(TARGET_OUT_GEN)", "STATIC_LIBRARIES", m.Name())
	}
	return filepath.Join("$(HOST_OUT_GEN)", "STATIC_LIBRARIES", m.Name())
}

func outputsVarName(m *GenerateCommon) string {
	return m.Name() + "_OUTPUTS"
}

func (g *AndroidMkGenerator) Outputs(m *GenerateCommon) string {
	return "$(" + outputsVarName(m) + ")"
}

const cmn_library_mk_text string = "include $(BUILD_SYSTEM)/base_rules.mk\n\n" +
	"export_includes:=$(intermediates)/export_includes\n" +

	//  Setup rule to create export_includes
	//  See prebuilt_internal.mk
	"$(export_includes): PRIVATE_EXPORT_C_INCLUDE_DIRS:=$(LOCAL_EXPORT_C_INCLUDE_DIRS)\n" +
	"$(export_includes): $(LOCAL_MODULE_MAKEFILE_DEP)\n" +
	"\t@echo Export includes file: $< -- $@\n" +
	"\t$(hide) mkdir -p $(dir $@) && rm -f $@\n" +
	"ifdef LOCAL_EXPORT_C_INCLUDE_DIRS\n" +
	"\t$(hide) for d in $(PRIVATE_EXPORT_C_INCLUDE_DIRS); do \\\n" +
	"\t\techo \"-I $$d\" >> $@; \\\n" +
	"\t\tdone\n" +
	"else\n" +
	"\t$(hide) touch $@\n" +
	"endif\n\n" +

	"$(LOCAL_BUILT_MODULE): $(LOCAL_SRC_FILES) | $(export_includes)\n" +
	"\tmkdir -p $(dir $@)\n" +
	"\tcp $< $@\n"

func declarePrebuiltStaticLib(moduleName, path, includePaths string, target bool) string {
	text := "\ninclude $(CLEAR_VARS)\n"
	text += "LOCAL_MODULE:=" + moduleName + "\n"
	text += "LOCAL_SRC_FILES:=" + path + "\n"
	if !target {
		text += "LOCAL_IS_HOST_MODULE:=true\n"
	}

	// We would like to just have the following line, but it looks like it is NDK only
	// Therefore all the following is needed.
	//text += "include $(PREBUILT_STATIC_LIBRARY)\n"

	text += "LOCAL_MODULE_CLASS:=STATIC_LIBRARIES\n"
	text += "LOCAL_UNINSTALLABLE_MODULE:=true\n"
	text += "LOCAL_MODULE_SUFFIX:=.a\n"

	if includePaths != "" {
		text += "LOCAL_EXPORT_C_INCLUDE_DIRS:=" + includePaths + "\n"
	}

	text += cmn_library_mk_text
	return text
}

func declarePrebuiltSharedLib(moduleName, path, includePaths string, target bool) string {
	text := "\ninclude $(CLEAR_VARS)\n"
	text += "LOCAL_MODULE:=" + moduleName + "\n"
	text += "LOCAL_SRC_FILES:=" + path + "\n"
	if !target {
		text += "LOCAL_IS_HOST_MODULE:=true\n"
	}
	// We would like to just have the following line, but it looks like it is NDK only
	// Therefore all the following is needed.
	//text += "include $(PREBUILT_SHARED_LIBRARY)\n"

	text += "LOCAL_MODULE_CLASS:=SHARED_LIBRARIES\n"
	text += "LOCAL_MODULE_SUFFIX:=.so\n"

	if includePaths != "" {
		text += "LOCAL_EXPORT_C_INCLUDE_DIRS:=" + includePaths + "\n"
	}

	//  Put shared libraries in common path to simplify link line.
	//  see shared_library_internal.mk and host_shared_library_internal.mk
	if target {
		// Only supporting the primary target architecture.
		// Others are TARGET_2ND_x
		text += "OVERRIDE_BUILT_MODULE_PATH:=$(TARGET_OUT_INTERMEDIATE_LIBRARIES)\n\n"
	} else {
		// Only supporting the primary host architecture.
		// Others are HOST_2ND_x, HOST_CROSS_x, HOST_CROSS_2ND_x
		text += "OVERRIDE_BUILT_MODULE_PATH:=$(HOST_OUT_INTERMEDIATE_LIBRARIES)\n\n"
	}

	text += cmn_library_mk_text
	return text
}

func declarePrebuiltBinary(moduleName, path string, target bool) string {
	text := "\ninclude $(CLEAR_VARS)\n"
	text += "LOCAL_MODULE:=" + moduleName + "\n"
	text += "LOCAL_SRC_FILES:=" + path + "\n"
	if !target {
		text += "LOCAL_IS_HOST_MODULE:=true\n"
	}

	text += "LOCAL_MODULE_CLASS:=EXECUTABLES\n"
	text += "LOCAL_MODULE_SUFFIX:=\n\n"

	text += "include $(BUILD_SYSTEM)/base_rules.mk\n\n"

	text += "$(LOCAL_BUILT_MODULE): $(LOCAL_SRC_FILES)\n"
	text += "\tmkdir -p $(dir $@)\n"
	text += "\tcp $< $@\n"

	return text
}

func installGeneratedFiles(m Installable, ctx blueprint.ModuleContext, tags []string) string {
	/* Install generated files one by one, if required */
	installGroupPath, ok := getInstallGroupPath(ctx)

	if !ok {
		return ""
	}

	text := "\n"
	filesToInstall := m.FilesToInstall(ctx)

	for _, file := range filesToInstall {
		moduleName := pathToModuleName(file)

		text += "include $(CLEAR_VARS)\n\n"

		text += "LOCAL_MODULE := " + moduleName + "\n"
		text += "LOCAL_INSTALLED_MODULE_STEM := " + filepath.Base(file) + "\n"
		text += "LOCAL_MODULE_CLASS := ETC\n"
		text += "LOCAL_MODULE_PATH := " + installGroupPath + "\n"
		text += "LOCAL_MODULE_RELATIVE_PATH := " + m.GetInstallableProps().Relative_install_path + "\n"
		text += "LOCAL_MODULE_TAGS := " + strings.Join(tags, " ") + "\n"
		text += "LOCAL_PREBUILT_MODULE_FILE := " + file + "\n\n"

		text += "include $(BUILD_PREBUILT)\n"
	}

	return text
}

func (g *AndroidMkGenerator) GenerateCommonActions(m *GenerateCommon, ctx blueprint.ModuleContext, inouts []Inout) string {
	text := "##########################\ninclude $(CLEAR_VARS)\n\n"

	// This is required to have $(local-generated-sources-dir) work as expected
	text += "LOCAL_MODULE := " + m.Name() + "\n"
	text += "LOCAL_MODULE_CLASS := STATIC_LIBRARIES\n"
	text += outputsVarName(m) + " := \n"
	text += outputDirVarName(m) + " := " + g.sourceOutputDir(m) + "\n"
	text += "\n"

	cmd, args, implicits, _ := m.GetArgs(ctx)
	utils.StripUnusedArgs(args, cmd)

	for _, inout := range inouts {
		if _, ok := args["headers_generated"]; ok {
			headers := utils.Filter(inout.out, utils.IsHeader)
			args["header_generated"] = strings.Join(headers, " ")
		}
		if _, ok := args["srcs_generated"]; ok {
			sources := utils.Filter(inout.out, utils.IsSource)
			args["srcs_generated"] = strings.Join(sources, " ")
		}

		in_arr := append(utils.PrefixDirs(inout.src_in, g.sourcePrefix()), inout.gen_in...)
		ins := strings.Join(in_arr, " ")

		// Make does not cleanly support multiple out-files
		// To handle that, we output the rule only on the first file, and let every other output
		// depend on the first.
		// This is not 100 % safe, since if the secondary file is removed, it will not be rebuilt.
		// It is assumed that this will not be a big issue, since removing individual files from a generated
		// directory should not be common.
		outs := inout.out[0]
		for _, key := range utils.SortedKeys(args) {
			text += outs + ": " + key + ":= " + args[key] + "\n"
		}

		text += outs + ": in := " + ins + "\n"
		text += outs + ": out := " + strings.Join(inout.out, " ") + "\n"
		text += outs + ": depfile := " + inout.depfile + "\n"
		text += outs + ": " + ins + " " + strings.Join(inout.implicit_srcs, " ") + "\n"
		text += "\t" + cmd + "\n"
		if inout.depfile != "" {
			// Convert the depfile file format as part of this rule.
			text += g.transformDepFile("$(depfile)")
			// ...and include it outside the rule.
			text += g.includeDepFile(outs, inout.depfile)
		}
		text += outputsVarName(m) + " += " + outs + "\n"

		for _, out := range inout.out[1:] {
			text += out + ": " + outs + "\n"
			text += outputsVarName(m) + " += " + out + "\n"
		}
		text += "\n"
	}
	text += g.Outputs(m) + ": " + strings.Join(implicits, " ") + "\n"

	/* This will ensure that any dependencies will not be rebuilt in the case of no change */
	text += ".KATI_RESTAT: " + g.Outputs(m) + "\n"

	return text
}

func (g *AndroidMkGenerator) GenerateSourceActions(m *GenerateSource, ctx blueprint.ModuleContext, inouts []Inout) {
	if enabledAndRequired(m) {
		text := g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)
		text += installGeneratedFiles(m, ctx, m.GenerateCommon.Properties.Tags)
		androidMkWriteString(ctx, m.AltShortName(), text)
	}
}

func (g *AndroidMkGenerator) TransformSourceActions(m *TransformSource, ctx blueprint.ModuleContext, inouts []Inout) {
	if enabledAndRequired(m) {
		text := g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)
		text += installGeneratedFiles(m, ctx, m.GenerateCommon.Properties.Tags)
		androidMkWriteString(ctx, m.AltShortName(), text)
	}
}

func (g *AndroidMkGenerator) GenStaticActions(m *GenerateStaticLibrary, ctx blueprint.ModuleContext, inouts []Inout) {
	if enabledAndRequired(m) {
		text := g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

		// Add prebuilt outputs
		include_dirs := utils.PrefixDirs(m.GenerateCommon.Properties.Export_gen_include_dirs, g.sourceOutputDir(&m.GenerateCommon))
		text += declarePrebuiltStaticLib(m.AltShortName(), getLibraryGeneratedPath(m, g),
			strings.Join(include_dirs, " "),
			m.GenerateCommon.Properties.Target != HOST_TAG)

		androidMkWriteString(ctx, m.AltShortName(), text)
	}
}

func (g *AndroidMkGenerator) GenSharedActions(m *GenerateSharedLibrary, ctx blueprint.ModuleContext, inouts []Inout) {
	if enabledAndRequired(m) {
		text := g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

		// Add prebuilt outputs
		include_dirs := utils.PrefixDirs(m.GenerateCommon.Properties.Export_gen_include_dirs, g.sourceOutputDir(&m.GenerateCommon))
		text += declarePrebuiltSharedLib(m.AltShortName(), getLibraryGeneratedPath(m, g),
			strings.Join(include_dirs, " "),
			m.GenerateCommon.Properties.Target != HOST_TAG)

		androidMkWriteString(ctx, m.AltShortName(), text)
	}
}

func (g *AndroidMkGenerator) GenBinaryActions(m *GenerateBinary, ctx blueprint.ModuleContext, inouts []Inout) {
	if enabledAndRequired(m) {
		text := g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

		// Add prebuilt outputs
		text += declarePrebuiltBinary(m.AltShortName(), getLibraryGeneratedPath(m, g),
			m.GenerateCommon.Properties.Target != HOST_TAG)

		androidMkWriteString(ctx, m.AltShortName(), text)
	}
}

func (g *AndroidMkGenerator) binaryOutputDir(m *Binary) string {
	return "$(HOST_OUT_EXECUTABLES)"
}

type AndroidMkOrderer struct {
}

type AndroidMkFile struct {
	Name string
	Deps []string
}

type AndroidMkFileSlice []AndroidMkFile

type AndroidModuleName interface {
	AltName() string
	AltShortName() string
}

func enabledAndRequired(m blueprint.Module) bool {
	if e, ok := m.(Enableable); ok {
		if !IsEnabled(e) || !IsRequired(e) {
			return false
		}
	}
	return true
}

func (s *AndroidMkOrderer) GenerateBuildActions(ctx blueprint.SingletonContext) {

	var order AndroidMkFileSlice
	ctx.VisitAllModules(func(m blueprint.Module) {
		di, ok := m.(AndroidModuleName)
		if ok && enabledAndRequired(m) {
			deps := []string{}
			ctx.VisitDepsDepthFirst(m, func(child blueprint.Module) {
				childdi, ok := child.(AndroidModuleName)
				_, is_defaults := child.(*Defaults)
				if ok && !is_defaults && enabledAndRequired(m) {
					deps = append(deps, childdi.AltShortName())
				}
			})
			_, is_defaults := di.(*Defaults)
			if !is_defaults {
				order = append(order, AndroidMkFile{di.AltShortName(), deps})
			}
		}
	})

	text := ""
	// Quick and dirty alphabetical topological sort
	for len(order) > 0 {
		lowindex := -1
		for i, val := range order {
			if (lowindex == -1 || val.Name < order[lowindex].Name) && len(val.Deps) == 0 {
				lowindex = i
			}
		}
		if lowindex == -1 {

			/* Generate a list of remaining modules and their dependencies */
			deps := ""
			for _, o := range order {
				deps += fmt.Sprintf("%s depends on\n", o.Name)
				for _, d := range o.Deps {
					deps += fmt.Sprintf("\t%s\n", d)
				}
			}

			panic(fmt.Errorf("unmet or circular dependency. %d remaining.\n%s", len(order), deps))
		}
		text += "include $(BOB_ANDROIDMK_DIR)/" + order[lowindex].Name + ".inc\n"
		for i := range order {
			newdeps := []string{}
			for _, dep := range order[i].Deps {
				if dep != order[lowindex].Name {
					newdeps = append(newdeps, dep)
				}
			}
			order[i].Deps = newdeps
		}
		order = append(order[:lowindex], order[lowindex+1:]...)
	}
	writeIfChanged(filepath.Join(builddir, "Android.inc"), text)
}

func (g *AndroidMkGenerator) moduleOutputDir(moduleName string) string {
	return fmt.Sprintf("$(dir $(ALL_MODULES.%s.BUILT))", moduleName)
}

func (g *AndroidMkGenerator) kernelModOutputDir(m *KernelModule) string {
	return g.moduleOutputDir(m.AltName())
}

func (g *AndroidMkGenerator) staticLibOutputDir(m *StaticLibrary) string {
	return g.moduleOutputDir(m.AltName())
}

func (g *AndroidMkGenerator) sharedLibOutputDir(m *SharedLibrary) string {
	return g.moduleOutputDir(m.AltName())
}

func (g *AndroidMkGenerator) sharedLibsDir(targetType string) string {
	if targetType != HOST_TAG {
		return "$(TARGET_OUT_SHARED_LIBRARIES)"
	}
	return "$(HOST_OUT_SHARED_LIBRARIES)"
}

func (g *AndroidMkGenerator) transformDepFile(depfile string) (text string) {
	text += "ifeq ($(word 1, $(subst ., ,$(PLATFORM_VERSION))),7)\n"
	text += "\t$(call transform-d-to-p-args," + depfile + "," + depfile + ".P)\n"
	text += "endif\n"
	return
}

func (g *AndroidMkGenerator) includeDepFile(target string, depfile string) (text string) {
	text += "ifeq ($(word 1, $(subst ., ,$(PLATFORM_VERSION))),7)\n"
	text += "  $(call include-depfile," + depfile + ".P," + target + ")\n"
	text += "else\n"
	text += "  $(call include-depfile," + depfile + "," + target + ")\n"
	text += "endif\n"
	return
}

func (g *AndroidMkGenerator) KernelModuleActions(m *KernelModule, ctx blueprint.ModuleContext) {
	if !enabledAndRequired(m) {
		return
	}

	text := "##########################\ninclude $(CLEAR_VARS)\n\n"
	text += "LOCAL_MODULE := " + m.AltShortName() + "\n"
	text += "LOCAL_MODULE_CLASS := KERNEL_MODULES\n"
	text += "LOCAL_CLANG := false\n"
	text += "LOCAL_MODULE_TAGS := " + strings.Join(m.Properties.Tags, " ") + "\n\n"

	sources := m.Properties.GetSrcs(ctx)

	text += "LOCAL_SRC_FILES :=" + newlineSeparatedList(sources)

	// Now the build rules (i.e. what would be done by an 'include
	// $(BUILD_KERNEL_MODULE)', if there was one).
	text += "TARGET_OUT_$(LOCAL_MODULE_CLASS) := $(TARGET_OUT)/lib/modules\n"
	installGroupPath, ok := getInstallGroupPath(ctx)
	if ok {
		text += "LOCAL_MODULE_PATH := " + installGroupPath + "\n"
		text += "LOCAL_MODULE_RELATIVE_PATH := " + m.Properties.Relative_install_path + "\n"
	} else {
		text += "LOCAL_UNINSTALLABLE_MODULE := true\n"
	}
	text += "LOCAL_MODULE_SUFFIX := .ko\n"
	if m.Properties.Owner != "" {
		text += "LOCAL_MODULE_OWNER := " + m.Properties.Owner + "\n"
		text += "LOCAL_PROPRIETARY_MODULE := true\n"
	}
	text += "include $(BUILD_SYSTEM)/base_rules.mk\n\n"

	args := m.generateKbuildArgs(ctx)
	args["sources"] = "$(addprefix $(LOCAL_PATH)/,$(LOCAL_SRC_FILES))"
	args["kmod_build"] = "$(LOCAL_PATH)/" + args["kmod_build"]
	args["local_path"] = "$(LOCAL_PATH)"

	// Create a target-specific variable declaration for each required parameter.
	for _, key := range utils.SortedKeys(args) {
		text += fmt.Sprintf("$(LOCAL_BUILT_MODULE): %s := %s\n", key, args[key])
	}

	text += "\n$(LOCAL_BUILT_MODULE): $(LOCAL_MODULE_MAKEFILE_DEP) " +
		args["sources"] + " " +
		args["kmod_build"] + " " +
		strings.Join(m.extraSymbolsFiles(ctx), " ") + "\n"
	text += "\tmkdir -p \"$(@D)\"\n"
	cmd := "python $(kmod_build) --output $@ --depfile $@.d " +
		"--common-root $(local_path) " +
		"--module-dir \"$(output_module_dir)\" $(extra_includes) " +
		"--sources $(sources) $(kbuild_extra_symbols) " +
		"--kernel \"$(kernel_dir)\" --cross-compile \"$(kernel_compiler)\" " +
		"$(kbuild_options) --extra-cflags \"$(extra_cflags)\" $(make_args)"

	text += "\techo " + cmd + "\n"
	text += "\t" + cmd + "\n"
	text += g.transformDepFile("$@.d") + "\n"

	text += g.includeDepFile("$(LOCAL_BUILT_MODULE)", "$(LOCAL_BUILT_MODULE).d")

	// The Module.symvers file is generated during Kbuild, but make doesn't
	// support multiple output files in a rule, so add it as a dependency
	// of the module.
	text += fmt.Sprintf("\n$(dir $(LOCAL_BUILT_MODULE))/Module.symvers: $(LOCAL_BUILT_MODULE)\n")

	androidMkWriteString(ctx, m.AltShortName(), text)
}

func AndroidMkOrdererFactory() blueprint.Singleton {
	return &AndroidMkOrderer{}
}

var (
	androidModuleNameMap    = map[string]string{}
	androidModuleReverseMap = map[string]string{}
	androidModuleMapLock    sync.RWMutex
)

func androidModuleName(name string) string {
	androidModuleMapLock.RLock()
	defer androidModuleMapLock.RUnlock()
	return androidModuleNameMap[name]
}

func mapAndroidNames(ctx blueprint.BottomUpMutatorContext) {
	if m, ok := ctx.Module().(AndroidModuleName); ok {
		// Ignore defaults
		if _, ok := ctx.Module().(*Defaults); ok {
			return
		}

		if enabledAndRequired(ctx.Module()) {
			androidModuleMapLock.Lock()
			defer androidModuleMapLock.Unlock()

			if existing, ok := androidModuleReverseMap[m.AltName()]; ok {
				if existing != ctx.ModuleName() {
					panic(fmt.Errorf("out name collision. Both %s and %s are required and map to %s",
						ctx.ModuleName(), existing, m.AltName()))
				}
			}
			androidModuleNameMap[ctx.ModuleName()] = m.AltName()
			androidModuleReverseMap[m.AltName()] = ctx.ModuleName()
		}
	}
}

func (*AndroidMkGenerator) Init(ctx *blueprint.Context) {
	ctx.RegisterBottomUpMutator("modulemapper", mapAndroidNames).Parallel()

	ctx.RegisterSingletonType("mali_androidmk_orderer", AndroidMkOrdererFactory)
}
