/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"errors"
	"os"
	"path/filepath"
	"reflect"
	"strings"

	"github.com/google/blueprint"
	"github.com/google/blueprint/bootstrap"
	"github.com/google/blueprint/proptools"

	"bob/graph"
	"bob/utils"
)

var (
	bobdir     = os.Getenv("BOB_DIR")
	srcdir     = os.Getenv("SRCDIR")
	builddir   = os.Getenv("BUILDDIR")
	configName = os.Getenv("CONFIGNAME")
	configOpts = os.Getenv("BOB_CONFIG_OPTS")
	configPath = filepath.Join(builddir, configName)
	jsonPath   = filepath.Join(builddir, "config.json")
)

// Module implementing this can support a phony target
type PhonyInterface interface {
	Outputs(g GeneratorBackend) []string
	ShortName() string
}

type ModuleWithBuildProps interface {
	Build() *Build
}

type TargetSpecific struct {
	Features
	BuildProps
}

// Modules implementing this can be depended upon
type DependentInterface interface {
	PhonyInterface
	OutputDir(g GeneratorBackend) string
}

type DependencyTag struct {
	blueprint.BaseDependencyTag
	name string
}

func getBackend(ctx blueprint.BaseModuleContext) GeneratorBackend {
	return getConfig(ctx).Generator
}

type GeneratorBackend interface {
	// Module build actions
	AliasActions(*Alias, blueprint.ModuleContext)
	BinaryActions(*Binary, blueprint.ModuleContext)
	GenerateSourceActions(*GenerateSource, blueprint.ModuleContext, []Inout)
	TransformSourceActions(*TransformSource, blueprint.ModuleContext, []Inout)
	GenSharedActions(*GenerateSharedLibrary, blueprint.ModuleContext, []Inout)
	GenStaticActions(*GenerateStaticLibrary, blueprint.ModuleContext, []Inout)
	GenBinaryActions(*GenerateBinary, blueprint.ModuleContext, []Inout)
	KernelModuleActions(m *KernelModule, ctx blueprint.ModuleContext)
	SharedActions(*SharedLibrary, blueprint.ModuleContext)
	StaticActions(*StaticLibrary, blueprint.ModuleContext)
	ResourceActions(*Resource, blueprint.ModuleContext)

	// Backend specific info for module types
	buildDir() string
	sourcePrefix() string
	sharedLibsDir(targetType string) string
	sourceOutputDir(m *GenerateCommon) string
	binaryOutputDir(m *Binary) string
	staticLibOutputDir(m *StaticLibrary) string
	sharedLibOutputDir(m *SharedLibrary) string
	kernelModOutputDir(m *KernelModule) string

	// Backend initialisation
	Init(*blueprint.Context)
}

type Config struct {
	Generator  GeneratorBackend
	Properties *ConfigProperties
}

type SourceProps struct {
	// The list of source files. Wildcards can be used (but are suboptimal)
	Srcs []string
	// The list of source files that should not be included. Use with care.
	Exclude_srcs []string

	// Sources that we need to treat specially
	Specials []string `blueprint:"mutated"`
}

func Glob(ctx blueprint.ModuleContext, globs []string, excludes []string) []string {
	var files []string

	for _, file := range globs {
		if strings.ContainsAny(file, "*?[") {
			matches, _ := ctx.GlobWithDeps(file, excludes)
			files = append(files, matches...)
		} else if !utils.Contains(excludes, file) {
			files = append(files, file)
		}
	}
	return files
}

// Get a list of sources to compile.
//
// The sources are relative to the project directory (i.e. include
// the module directory but not the base source directory), and
// excludes have been handled.
func (s *SourceProps) GetSrcs(ctx blueprint.ModuleContext) []string {
	return Glob(ctx, s.Srcs, s.Exclude_srcs)
}

func (s *SourceProps) processPaths(ctx blueprint.BaseModuleContext) {
	g := getBackend(ctx)
	prefix := ctx.ModuleDir()
	var special = map[string]string{
		"${bob_config}": filepath.Join(g.buildDir(), configName),
	}

	// Look for special items. Remove from Srcs and add to Specials
	srcs := []string{}
	for _, src := range s.Srcs {
		if value, ok := special[src]; !ok {
			srcs = append(srcs, src)
		} else {
			// Only append if not in Excluded.
			// Users shouldn't rely on how any special is expanded, so
			// no need to check the expanded case.
			if !utils.Contains(s.Exclude_srcs, src) {
				s.Specials = append(s.Specials, value)
			}
		}
	}

	s.Srcs = utils.PrefixDirs(srcs, prefix)
	s.Exclude_srcs = utils.PrefixDirs(s.Exclude_srcs, prefix)
}

type ConfigProvider interface {
	Config() interface{}
}

func getConfig(ctx ConfigProvider) *Config {
	return ctx.Config().(*Config)
}

type DependencySingleton struct{}

func (m *DependencySingleton) GenerateBuildActions(ctx blueprint.SingletonContext) {
	ctx.AddNinjaFileDeps(jsonPath)
}

func DependencySingletonFactory() blueprint.Singleton {
	return &DependencySingleton{}
}

const (
	HOST_TAG   string = "host"
	TARGET_TAG string = "target"
)

func stripEmptyComponentsRecursive(propsVal reflect.Value) {
	var empty_str_filter = func(s string) bool { return s != "" }

	for i := 0; i < propsVal.NumField(); i++ {
		field := propsVal.Field(i)

		switch field.Kind() {
		case reflect.Slice:
			if field.Type().Elem().Kind() == reflect.String {
				list := field.Interface().([]string)
				list = utils.Filter(list, empty_str_filter)
				field.Set(reflect.ValueOf(list))
			}

		case reflect.Struct:
			stripEmptyComponentsRecursive(field)
		}
	}
}

func StripEmptyComponentsMutator(mctx blueprint.BottomUpMutatorContext) {
	f, ok := mctx.Module().(Featurable)
	if !ok {
		return
	}

	for _, props := range f.topLevelProperties() {
		propsVal := reflect.Indirect(reflect.ValueOf(props))
		stripEmptyComponentsRecursive(propsVal)
	}
}

func moduleNamesFromLibList(libList []string) (ret []string) {
	moduleNames := make(map[string]bool)

	for _, lib := range libList {
		module, _ := splitGeneratedComponent(lib)
		if _, ok := moduleNames[module]; !ok {
			ret = append(ret, module)
			moduleNames[module] = true
		}
	}

	return ret
}

const splitterMutatorName string = "library"

func parseAndAddVariationDeps(mctx blueprint.BottomUpMutatorContext,
	tag blueprint.DependencyTag, deps ...string) {

	hostVariation := []blueprint.Variation{blueprint.Variation{Mutator: splitterMutatorName, Variation: HOST_TAG}}
	targetVariation := []blueprint.Variation{blueprint.Variation{Mutator: splitterMutatorName, Variation: TARGET_TAG}}

	for _, dep := range deps {
		var variations []blueprint.Variation

		idx := strings.LastIndex(dep, ":")
		if idx > 0 {
			variationNames := strings.Split(dep[idx+1:len(dep)], ",")
			for _, vn := range variationNames {
				if vn == "host" {
					variations = append(variations, hostVariation...)
				} else if vn == "target" {
					variations = append(variations, targetVariation...)
				} else {
					panic("Invalid variation: " + vn + " in module name " + dep)
				}
			}

			dep = dep[0:idx]
		}

		if len(variations) > 0 {
			mctx.AddVariationDependencies(variations, tag, dep)
		} else {
			mctx.AddDependency(mctx.Module(), tag, dep)
		}
	}
}

var WholeStaticDepTag = DependencyTag{name: "whole_static"}
var StaticDepTag = DependencyTag{name: "static"}
var SharedDepTag = DependencyTag{name: "shared"}
var FlagDepTag = DependencyTag{name: "reexport"}
var KernelModuleDepTag = DependencyTag{name: "kernel_module"}

type Targetable interface {
	Build() *Build
	Features() *Features
	getTarget() string
}

func DependerMutator(mctx blueprint.BottomUpMutatorContext) {
	if e, ok := mctx.Module().(Enableable); ok {
		if !IsEnabled(e) {
			// Not enabled, so don't add dependencies
			return
		}
	}

	if t, ok := mctx.Module().(Targetable); ok {
		build := t.Build()
		if _, ok := mctx.Module().(*Defaults); ok {
			// We do not want to add dependencies for defaults
			return
		}
		mctx.AddVariationDependencies(nil, WholeStaticDepTag, build.Whole_static_libs...)
		mctx.AddVariationDependencies(nil, StaticDepTag, build.Static_libs...)
		mctx.AddVariationDependencies(nil, StaticDepTag, build.Export_static_libs...)

		mctx.AddVariationDependencies(nil, SharedDepTag, build.Shared_libs...)
		mctx.AddVariationDependencies(nil, SharedDepTag, build.Export_shared_libs...)
	}
	if km, ok := mctx.Module().(*KernelModule); ok {
		mctx.AddDependency(mctx.Module(), KernelModuleDepTag, km.Properties.Extra_symbols...)
	}
	if ins, ok := mctx.Module().(Installable); ok {
		props := ins.GetInstallableProps()
		if props.Install_group != nil {
			mctx.AddDependency(mctx.Module(), InstallGroupTag, *props.Install_group)
		}
		parseAndAddVariationDeps(mctx, InstallDepTag, props.Install_deps...)
	}
}

// Applies target specific properties within each module. Must be done
// after the libraries have been split.
func TargetMutator(mctx blueprint.TopDownMutatorContext) {
	var build *Build
	var target string

	if def, ok := mctx.Module().(Targetable); ok {
		build = def.Build()
		target = def.getTarget()
	} else if gsc, ok := GetGenerateCommon(mctx.Module()); ok {
		build = &gsc.Properties.FlagArgsBuild
		target = gsc.Properties.Target
	} else {
		return
	}

	//print(mctx.ModuleName() + " is targetable\n")
	var src *TargetSpecific
	if target == HOST_TAG {
		src = &build.Host
	} else if target == TARGET_TAG {
		src = &build.Target
	} else {
		// This is fine - it can happen if the target is the default
		return
	}

	// Copy the target-specific variables to the core set
	err := proptools.AppendMatchingProperties([]interface{}{&build.BuildProps}, &src.BuildProps, nil)
	if err != nil {
		if propertyErr, ok := err.(*proptools.ExtendPropertyError); ok {
			mctx.PropertyErrorf(propertyErr.Property, "%s", propertyErr.Err.Error())
		} else {
			panic(err)
		}
	}
}

type pathProcessor interface {
	processPaths(blueprint.BaseModuleContext)
}

// Adds module paths to appropriate properties.
func PathMutator(mctx blueprint.BottomUpMutatorContext) {
	if p, ok := mctx.Module().(pathProcessor); ok {
		p.processPaths(mctx)
	}
}

func CollectReexportDependenciesMutator(mctx blueprint.TopDownMutatorContext) {
	mainModule := mctx.Module()
	if e, ok := mainModule.(Enableable); ok {
		if !IsEnabled(e) {
			return // Not enabled, so don't add dependencies
		}
	}

	var mainBuild *Build
	if buildProps, ok := mainModule.(ModuleWithBuildProps); ok {
		mainBuild = buildProps.Build()
	} else {
		return // We do not want to add dependencies for not targets
	}

	mctx.WalkDeps(func(child blueprint.Module, parent blueprint.Module) bool {
		depTag := mctx.OtherModuleDependencyTag(child)
		if depTag == WholeStaticDepTag || depTag == StaticDepTag || depTag == SharedDepTag {
			var parentBuild *Build
			if moduleWithBuildProps, ok := parent.(ModuleWithBuildProps); ok {
				parentBuild = moduleWithBuildProps.Build()
			} else {
				return false
			}

			var childBuild *Build
			if moduleWithBuildProps, ok := child.(ModuleWithBuildProps); ok {
				childBuild = moduleWithBuildProps.Build()
			} else {
				return false
			}

			if len(childBuild.Reexport_libs) > 0 &&
				(parent.Name() == mainModule.Name() || utils.Contains(parentBuild.Reexport_libs, child.Name())) {
				mainBuild.Resolved_reexported_libs = utils.AppendUnique(mainBuild.Resolved_reexported_libs, childBuild.Reexport_libs)
				return true
			}
		}

		return false
	})
}

func ApplyReexportDependenciesMutator(mctx blueprint.BottomUpMutatorContext) {
	mainModule := mctx.Module()
	if e, ok := mainModule.(Enableable); ok {
		if !IsEnabled(e) {
			// Not enabled, so don't add dependencies
			return
		}
	}

	var build *Build
	if buildProps, ok := mainModule.(ModuleWithBuildProps); ok {
		build = buildProps.Build()
		mctx.AddVariationDependencies(nil, FlagDepTag, build.Resolved_reexported_libs...)
	}
}

// Create a closure passing the config to a module factory so that the module
// factories can access the config.
type factoryWithConfig func(*Config) (blueprint.Module, []interface{})

func passConfig(mf factoryWithConfig, config *Config) blueprint.ModuleFactory {
	return func() (blueprint.Module, []interface{}) {
		return mf(config)
	}
}

func FindRequiredModulesMutator(ctx blueprint.TopDownMutatorContext) {
	// Non-enableable module types are aliases and defaults. All
	// dependencies of an alias should be required. Ignore defaults,
	// because they've already been applied and don't generate any build
	// rules themselves.
	if e, ok := ctx.Module().(Enableable); ok {
		// If it's a top-level module (enabled and built by default), mark it as
		// required, and continue to visit its dependencies. Otherwise, we don't
		// need its dependencies so return.
		if IsEnabled(e) && IsBuiltByDefault(e) {
			MarkAsRequired(e)
		} else {
			return
		}
	} else if _, ok := ctx.Module().(*Defaults); ok { // Ignore defaults.
		return
	} else if _, ok := ctx.Module().(*Alias); ok { // Ignore aliases.
		return
	}

	ctx.WalkDeps(func(dep blueprint.Module, parent blueprint.Module) bool {
		e, ok := dep.(Enableable)
		if ok {
			// Stop traversing if we've already visited this while
			// following another module's dependencies.
			if IsRequired(e) {
				return false
			}
			// Don't require disabled dependencies (for example aliases with
			// some disabled sources).
			if !IsEnabled(e) {
				return false
			}
			MarkAsRequired(e)
		}
		return true
	})
}

func Main() {
	// Load the config first. This is needed because some of the module
	// types' definitions contain a struct-per-feature, and features are
	// specified in the config.
	config := &Config{}
	config.Properties = LoadConfig(jsonPath)

	var ctx = blueprint.NewContext()
	ctx.RegisterModuleType("bob_binary", passConfig(BinaryFactory, config))
	ctx.RegisterModuleType("bob_static_library", passConfig(StaticLibraryFactory, config))
	ctx.RegisterModuleType("bob_shared_library", passConfig(SharedLibraryFactory, config))
	ctx.RegisterModuleType("bob_defaults", passConfig(DefaultsFactory, config))
	ctx.RegisterModuleType("bob_generate_source", passConfig(GenerateSourceFactory, config))
	ctx.RegisterModuleType("bob_transform_source", passConfig(TransformSourceFactory, config))
	ctx.RegisterModuleType("bob_generate_static_library", passConfig(genStaticLibFactory, config))
	ctx.RegisterModuleType("bob_generate_shared_library", passConfig(genSharedLibFactory, config))
	ctx.RegisterModuleType("bob_generate_binary", passConfig(genBinaryFactory, config))
	ctx.RegisterModuleType("bob_alias", passConfig(aliasFactory, config))
	ctx.RegisterModuleType("bob_kernel_module", passConfig(KernelModuleFactory, config))
	ctx.RegisterModuleType("bob_resource", passConfig(ResourceFactory, config))
	ctx.RegisterModuleType("bob_install_group", passConfig(InstallGroupFactory, config))

	// Note that the order of mutators are important, since the
	// contents of each module will be rewritten. The following
	// describes the required orderring of mutators dealing with
	// property propagation.
	//
	// Merge feature specific values to the level above in each
	// module. This must be before defaults so that a feature specific
	// option set in a default does not override an option set in a
	// module. Do this before templates so templates only need to
	// operate on one level.
	//
	// Evaluate templates next, including in defaults. This avoids us
	// having to re-evaluate templates after they have been copied
	// around by defaults.
	//
	// Then apply defaults. Do this before the library splitter so that
	// we can propagate target_supported and host_supported through
	// defaults if needed.
	//
	// Next split libraries into host and target specific modules.
	//
	// After the libraries are split we can apply target specific options.
	//
	// The depender mutator adds the dependencies between binaries and libraries.
	//
	// The generated depender mutator add dependencies to generated source modules.
	ctx.RegisterBottomUpMutator("default_deps", DefaultDepsMutator).Parallel()
	ctx.RegisterTopDownMutator("features_applier", FeatureApplierMutator).Parallel()
	ctx.RegisterTopDownMutator("template_applier", TemplateApplierMutator).Parallel()
	ctx.RegisterBottomUpMutator("strip_empty_components", StripEmptyComponentsMutator).Parallel()
	ctx.RegisterBottomUpMutator("process_paths", PathMutator).Parallel()
	ctx.RegisterTopDownMutator("supported_variants", SupportedVariantsMutator).Parallel()
	ctx.RegisterBottomUpMutator(splitterMutatorName, splitterMutator).Parallel()
	ctx.RegisterTopDownMutator("target", TargetMutator).Parallel()
	ctx.RegisterTopDownMutator("default_applier", DefaultApplierMutator).Parallel()
	ctx.RegisterBottomUpMutator("depender", DependerMutator).Parallel()
	ctx.RegisterBottomUpMutator("alias", aliasMutator).Parallel()
	ctx.RegisterBottomUpMutator("generated", generatedDependerMutator).Parallel()

	if handler := initGrapvizHandler(); handler != nil {
		ctx.RegisterBottomUpMutator("graphviz_output", handler.graphvizMutator)
		// Singleton for stop tool and don't overwrite build.bp
		ctx.RegisterSingletonType("quit_singleton", handler.quitSingletonFactory)
	} else {

		ctx.RegisterTopDownMutator("export_lib_flags", ExportLibFlagsMutator).Parallel()
		dependencyGraphHandler := graphMutatorHandler{graph.NewGraph("All")}
		ctx.RegisterBottomUpMutator("sort_resolved_static_libs", dependencyGraphHandler.ResolveDependencySortMutator) // This can't be parallel

		ctx.RegisterTopDownMutator("find_required_modules", FindRequiredModulesMutator).Parallel()

		ctx.RegisterTopDownMutator("collect_reexport_dependencies", CollectReexportDependenciesMutator).Parallel()
		ctx.RegisterBottomUpMutator("apply_reexport_dependencies", ApplyReexportDependenciesMutator).Parallel()

		// Depend on the config file
		ctx.RegisterSingletonType("config_singleton", DependencySingletonFactory)
	}

	if config.Properties.GetBool("builder_linux") {
		config.Generator = &LinuxGenerator{}
	} else if config.Properties.GetBool("builder_android") {
		config.Generator = &AndroidMkGenerator{}
	} else {
		panic(errors.New("unknown builder backend"))
	}

	config.Generator.Init(ctx)
	bootstrap.Main(ctx, config)
}
