/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"fmt"
	"reflect"

	"github.com/google/blueprint"
)

type EnableableProps struct {
	// Used to disable the generation of build rules. If this is set to false, no build rule will be generated.
	Enabled *bool
	// Whether it is built by default in a build with no targets requested.
	// Nothing to do with 'defaults'.
	Build_by_default *bool
	// Is this module depended on by a module which is built by default?
	// Used to prune unused modules from Android builds, where we can't
	// control exactly what gets built.
	Required bool `blueprint:"mutated"`
}

type Enableable interface {
	GetEnableableProps() *EnableableProps
}

func IsEnabled(e Enableable) bool {
	props := e.GetEnableableProps()
	if props.Enabled != nil {
		return *props.Enabled
	}
	return true
}

func IsBuiltByDefault(e Enableable) bool {
	props := e.GetEnableableProps()
	if props.Build_by_default != nil {
		return *props.Build_by_default
	}

	switch m := e.(type) {
	case *Binary:
		if m.Properties.Target_type == TARGET_TAG {
			return true
		}
	case *KernelModule:
		return true
	}
	return false
}

func IsRequired(e Enableable) bool {
	return e.GetEnableableProps().Required
}

func MarkAsRequired(e Enableable) {
	e.GetEnableableProps().Required = true
}

type InstallableProps struct {
	// Module specifying an installation directory
	Install_group *string
	// Other modules which must be installed alongside this
	Install_deps []string
	// Path to install to, relative to the install_group's path
	Relative_install_path string
	// Script used during post install
	Post_install_tool string
	// Command to execute on file(s) after they are installed
	Post_install_cmd string
}

func GetShortNamesForDirectDepsWithTags(ctx blueprint.ModuleContext,
	tags ...DependencyTag) (ret []string) {
	visited := map[string]bool{}
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool {
			tag := ctx.OtherModuleDependencyTag(m)
			for _, i := range tags {
				if tag == i {
					return true
				}
			}
			return false
		},
		func(m blueprint.Module) {
			if dep, ok := m.(PhonyInterface); ok {
				if _, ok := visited[m.Name()]; !ok {
					ret = append(ret, dep.ShortName())
				}
			} else {
				panic("install_dep on non-dependendable module")
			}
		})
	return
}

type InstallGroupProps struct {
	Install_path string
}

type InstallGroup struct {
	blueprint.SimpleName
	Properties struct {
		InstallGroupProps
		Features
	}
}

func (m *InstallGroup) GenerateBuildActions(ctx blueprint.ModuleContext) {
	// No build actions for a bob_install_group
}

func (m *InstallGroup) topLevelProperties() []interface{} {
	return []interface{}{&m.Properties.InstallGroupProps}
}

func (m *InstallGroup) Features() *Features {
	return &m.Properties.Features
}

type SymlinkInstaller interface {
	LibrarySymlinks(ctx blueprint.ModuleContext) map[string]string
}

type Installable interface {
	FilesToInstall(ctx blueprint.ModuleContext) []string
	GetInstallableProps() *InstallableProps
	GetInstallDepPhonyNames(ctx blueprint.ModuleContext) []string
}

type ResourceProps struct {
	SourceProps
	AliasableProps
	InstallableProps
	EnableableProps
	Tags []string
}

type Resource struct {
	blueprint.SimpleName
	Properties struct {
		ResourceProps
		Features
	}
}

func getInstallGroupPath(ctx blueprint.ModuleContext) (string, bool) {
	var installGroupPath *string

	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool {
			return ctx.OtherModuleDependencyTag(m) == InstallGroupTag
		},
		func(m blueprint.Module) {
			insg, ok := m.(*InstallGroup)
			if !ok {
				panic(fmt.Sprintf("install_group dependency of %s not an install group",
					ctx.ModuleName()))
			}
			if installGroupPath != nil {
				panic(fmt.Sprintf("Multiple install group dependencies for %s",
					ctx.ModuleName()))
			}
			installGroupPath = &insg.Properties.Install_path
		})

	if installGroupPath == nil {
		return "", false
	}

	if *installGroupPath == "" {
		panic(fmt.Sprintf("Module %s has empty install path", ctx.ModuleName()))
	}

	return *installGroupPath, true
}

func (m *Resource) GenerateBuildActions(ctx blueprint.ModuleContext) {
	if IsEnabled(m) {
		getBackend(ctx).ResourceActions(m, ctx)
	}
}

func (m *Resource) topLevelProperties() []interface{} {
	return []interface{}{&m.Properties.ResourceProps}
}

func (m *Resource) Features() *Features {
	return &m.Properties.Features
}

func (m *Resource) GetInstallDepPhonyNames(ctx blueprint.ModuleContext) []string {
	return GetShortNamesForDirectDepsWithTags(ctx, InstallDepTag)
}

func (m *Resource) ShortName() string {
	return m.Name()
}

func (m *Resource) AltName() string {
	return m.Name()
}

func (m *Resource) AltShortName() string {
	return m.ShortName()
}

func (m *Resource) GetEnableableProps() *EnableableProps {
	return &m.Properties.EnableableProps
}

// Resources don't have any outputs (i.e. stuff generated in the build
// directory) - they only copy source files to the installation dir. This
// method exists to implement PhonyInterface.
func (m *Resource) Outputs(g GeneratorBackend) []string {
	return []string{}
}

func (m *Resource) FilesToInstall(ctx blueprint.ModuleContext) []string {
	return m.Properties.SourceProps.GetSrcs(ctx)
}

func (m *Resource) GetInstallableProps() *InstallableProps {
	return &m.Properties.InstallableProps
}

func (m *Resource) processPaths(ctx blueprint.BaseModuleContext) {
	m.Properties.SourceProps.processPaths(ctx)
}

func (m *Resource) getAliasList() []string {
	return m.Properties.getAliasList()
}

func InstallGroupFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &InstallGroup{}
	module.Properties.Features.Init(config, reflect.TypeOf(InstallGroupProps{}))
	return module, []interface{}{&module.Properties,
		&module.SimpleName.Properties}
}

func ResourceFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &Resource{}
	module.Properties.Features.Init(config, reflect.TypeOf(ResourceProps{}))
	return module, []interface{}{&module.Properties,
		&module.SimpleName.Properties}
}

var InstallGroupTag = DependencyTag{name: "install_group"}
var InstallDepTag = DependencyTag{name: "install_dep"}
