/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"fmt"

	"github.com/google/blueprint"
	"github.com/google/blueprint/proptools"
)

// Used to map a set of properties to destination properties
type propmap struct {
	dst []interface{}
	src interface{}
}

// Applies default options
func DefaultApplierMutator(mctx blueprint.TopDownMutatorContext) {
	// This method walks down the dependency list to include all defaults that include other defaults
	// with the ones further down the tree being applied first.
	// Walkdeps is a preorder depth-first search - meaning a parent is visited before children, and children
	// is visited before siblings.
	_, isDefaults := mctx.Module().(*Defaults)
	if isDefaults {
		return
	}

	var build *Build

	if target, ok := mctx.Module().(Defaultable); ok {
		build = target.Build()
	} else if gsc, ok := GetGenerateCommon(mctx.Module()); ok {
		build = &gsc.Properties.FlagArgsBuild
	} else {
		// Not defaultable.
		return
	}

	visited := map[string]bool{}

	mctx.WalkDeps(func(dep blueprint.Module, parent blueprint.Module) bool {
		if mctx.OtherModuleDependencyTag(dep) == defaultDepTag {
			//print("Visiting " + mctx.OtherModuleName(dep) + " for dependency " + mctx.ModuleName() + "\n")
			def, ok := dep.(*Defaults)
			if !ok {
				panic(fmt.Errorf("module %s in %s's defaults is not a default",
					dep.Name(), mctx.ModuleName()))
			}

			// Only visit each default once
			if _, ok := visited[dep.Name()]; ok {
				return false
			}
			visited[dep.Name()] = true

			// Defaults are more generic, so we prepend to the
			// core module properties.
			//
			// Note: when prepending (pointers to) bools we copy
			// the value if the dst is nil, otherwise the dst
			// value is left alone.
			err := proptools.PrependMatchingProperties([]interface{}{&build.BuildProps}, &def.Build().BuildProps, nil)
			if err != nil {
				if propertyErr, ok := err.(*proptools.ExtendPropertyError); ok {
					mctx.PropertyErrorf(propertyErr.Property, "%s", propertyErr.Err.Error())
				} else {
					panic(err)
				}
			}

			return true // This return value indicates if we want to continue visiting children.
		}
		return false
	})
}

// This module type supports the use of features and templates.
type Featurable interface {
	topLevelProperties() []interface{}
	Features() *Features
}

func TemplateApplierMutator(mctx blueprint.TopDownMutatorContext) {
	if m, ok := mctx.Module().(Featurable); ok {
		cfgProps := getConfig(mctx).Properties

		// TemplateApplier mutator is run before TargetApplier, so we
		// need to apply templates with the core set, as well as
		// host-specific and target-specific sets (where applicable).
		props := append([]interface{}{}, m.topLevelProperties()...)
		if m, ok := mctx.Module().(ModuleWithBuildProps); ok {
			props = append(props, &m.Build().Host.BuildProps)
			props = append(props, &m.Build().Target.BuildProps)
		}
		for _, p := range props {
			ApplyTemplate(p, cfgProps)
		}
	}
}

// Applies feature specific properties within each module
func FeatureApplierMutator(mctx blueprint.TopDownMutatorContext) {
	if m, ok := mctx.Module().(Featurable); ok {
		cfgProps := getConfig(mctx).Properties

		// FeatureApplier mutator is run first. We need to flatten the
		// feature specific properties in the core set, and where
		// supported, the host-specific and target-specific set.
		var props = []propmap{propmap{m.topLevelProperties(), m.Features().BlueprintEmbed}}

		if m, ok := mctx.Module().(ModuleWithBuildProps); ok {
			var tgtprops = []propmap{
				propmap{[]interface{}{&m.Build().Host.BuildProps}, m.Build().Host.Features.BlueprintEmbed},
				propmap{[]interface{}{&m.Build().Target.BuildProps}, m.Build().Target.Features.BlueprintEmbed},
			}
			props = append(props, tgtprops...)
		}

		for _, prop := range props {
			// Feature specific properties get added after core properties.
			//
			// Note: when appending (pointers to) bools we always override
			// the dst value. i.e. feature-specific value takes precedence.
			err := AppendProps(prop.dst, prop.src, cfgProps)
			if err != nil {
				if propertyErr, ok := err.(*proptools.ExtendPropertyError); ok {
					mctx.PropertyErrorf(propertyErr.Property, "%s", propertyErr.Err.Error())
				} else {
					panic(err)
				}
			}
		}
	}
}
