/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */
package core

import (
	"errors"
	"fmt"

	"github.com/google/blueprint"
	"github.com/google/blueprint/proptools"
)

type SplittableProps struct {
	Host_supported   *bool
	Target_supported *bool
}

// If a Module implements this interface, then it will be split into
// the different variants by the splitterMutator
type splittable interface {
	// Retrieve all the different variations to create
	supportedVariants() []string

	// Disables the module is no variations supported
	disable()

	// Set the particular variant
	setVariant(string)

	// Get the properties related to which variants are available
	getSplittableProps() *SplittableProps
}

// Traverse the core properties of defaults to find out which variations are
// supported.
func SupportedVariantsMutator(mctx blueprint.TopDownMutatorContext) {
	sp, ok := mctx.Module().(splittable)
	if !ok {
		return
	}

	// It's not valid to specify host_supported or target_supported in a
	// target: {} or host: {} section
	if moduleWithBuildProps, ok := mctx.Module().(ModuleWithBuildProps); ok {
		b := moduleWithBuildProps.Build()
		if b.Host.SplittableProps.Host_supported != nil ||
			b.Host.SplittableProps.Target_supported != nil ||
			b.Target.SplittableProps.Host_supported != nil ||
			b.Target.SplittableProps.Target_supported != nil {
			panic(fmt.Errorf("target-specific [host|target]_supported in module %s",
				mctx.ModuleName()))
		}
	}

	// Defaults are always split into both variants
	if _, isDefaults := mctx.Module().(*Defaults); isDefaults {
		return
	}

	visited := map[string]bool{}

	mctx.WalkDeps(func(dep blueprint.Module, parent blueprint.Module) bool {
		if mctx.OtherModuleDependencyTag(dep) == defaultDepTag {
			def, ok := dep.(*Defaults)
			if !ok {
				panic(fmt.Errorf("module %s in %s's defaults is not a default",
					dep.Name(), mctx.ModuleName()))
			}

			// Only visit each default once
			if _, ok := visited[dep.Name()]; ok {
				return false
			}
			visited[dep.Name()] = true

			err := proptools.PrependMatchingProperties([]interface{}{sp.getSplittableProps()},
				&def.Properties.SplittableProps, nil)
			if err != nil {
				if propertyErr, ok := err.(*proptools.ExtendPropertyError); ok {
					mctx.PropertyErrorf(propertyErr.Property, "%s", propertyErr.Err.Error())
				} else {
					panic(err)
				}
			}
			return true
		}
		return false
	})
}

// Creates all the supported variants of splittable modules, including defaults.
func splitterMutator(mctx blueprint.BottomUpMutatorContext) {
	if s, ok := mctx.Module().(splittable); ok {
		variants := s.supportedVariants()
		if len(variants) == 0 {
			s.disable()
		} else {
			modules := mctx.CreateVariations(variants...)
			for i, v := range variants {
				newsplit, ok := modules[i].(splittable)
				if !ok {
					panic(errors.New("newly created variation is not splittable - should not happen"))
				}
				newsplit.setVariant(v)
			}
		}
	}
}
